/*
 * file: NrSpeedBuf.cpp
 *
 * This file is part of the NetStreamer software. This file is distributed
 * under the GNU GENERAL PUBLIC LICENSE, see the accompanying COPYING file.
 *
 * Copyright (C) 1997 Rolf Fokkens
 *
 */

#include <limits.h>

#include "NrSpeedBuf.h"

inline void AdjustAvg (int &Avg, int New, int Weight)
{
    Avg = ((Avg << Weight) - Avg + New ) >> Weight;
};

int NrSpeedBuf::GetPutSpeed (int shift)
{
    return PutCalc.GetAvgSpeed (shift);
};

int NrSpeedBuf::GetGetSpeed (int shift)
{
    return GetCalc.GetAvgSpeed (shift);
};

void NrSpeedBuf::IntTurnOff (void)
{
    if (!MuteFlag && OnFlag) return;

    if (State == Off) return;

    if (State == Preparing) HandlePreparing (0);

    State = Off;

    NrFifo::Reset ();
};

void NrSpeedBuf::IntTurnOn (void)
{
    if (MuteFlag || !OnFlag) return;

    PutCalc.Reset ();
    GetCalc.Reset ();

    AvgPutSpeed = 32000 << 8;
    AvgGetSpeed = 32000 << 8;

    AvgCounter  = 0;
    AvgDivisor  = 1;

    if (State != Preparing) HandlePreparing (1);
    State = Preparing;
};

void NrSpeedBuf::TurnOff (void)
{
    if (!OnFlag) return;

    OnFlag = 0;

    IntTurnOff ();
};

void NrSpeedBuf::TurnOn (void)
{
    if (OnFlag) return;

    OnFlag = 1;

    IntTurnOn ();
};

void NrSpeedBuf::SetMute (int Flag)
{
    if (MuteFlag == Flag) return;

    MuteFlag = Flag;

    if (MuteFlag) {
        IntTurnOff ();
    } else {
        IntTurnOn ();
    }
};

void NrSpeedBuf::HandlePreparing (int Flag)
{
};

NrSpeedBuf::NrSpeedBuf (int AimSize)
    : NrFifo (AimSize << 1)
    , Decomp (3)
{
    NrSpeedBuf::SpeedCount  = 0;
    NrSpeedBuf::ExpCount    = 0;

    NrSpeedBuf::AimSize     = AimSize;
    NrSpeedBuf::AvgAimVal   = 4000;

    NrSpeedBuf::AvgPutSpeed = 32000 << 8;
    NrSpeedBuf::AvgGetSpeed = 32000 << 8;

    NrSpeedBuf::AvgCounter  = 0;
    NrSpeedBuf::AvgDivisor  = 1;

    NrSpeedBuf::MuteFlag    = 0;
    NrSpeedBuf::OnFlag      = 0;

    IntTurnOff ();
};

NrSpeedBuf::~NrSpeedBuf (int AimSize)
{
};

void NrSpeedBuf::PutSound (EzString Sound, long MsCurTime, int Expansion)
{
    int TmpPutSpeed;
    int SoundSize;

    if (State == Off) return;

    AdjustAvg (AvgBufSize, GetBufUsed (), 10);

    SoundSize = NrFifo::PutSound
                    (Decomp.Decompress (Sound, AUDIO_BLOCK), Expansion);

    HistBuf.PutData (Sound, SoundSize);

    PutCalc.AddData (SoundSize * 2, MsCurTime);

    TmpPutSpeed = (PutCalc.GetAvgSpeed (8) * 1000);

    AdjustAvg (AvgPutSpeed, TmpPutSpeed, 6);

    if (State == Preparing && GetBufUsed () > (GetBufSize () >> 1)) {
        HandlePreparing (0);
        State       = On;
        AvgPutSpeed = PutCalc.GetAvgSpeed (8) * 1000;
        AvgGetSpeed = GetCalc.GetAvgSpeed (8) * 1000;
    };
};

static void ReduceFract
    ( int &Counter1, int &Divisor1
    )
{
    int C, D, S;

    if (Divisor1 < 0) {
        C = -Counter1;
        D = -Divisor1;
    } else {
        C =  Counter1;
        D =  Divisor1;
    };
    S = C < 0;

    if (S != 0) C = -C;

    while (C > 0x08000 || D > 0x08000) {
        C >>= 1;
        D >>= 1;
    };
    if (S !=0) C = -C;

    Counter1 = C;
    Divisor1 = D;
};

static void MultFract
    ( int Counter1,  int Divisor1
    , int Counter2,  int Divisor2
    , int &Counter3, int &Divisor3
    )
{
    ReduceFract (Counter1, Divisor1);
    ReduceFract (Counter2, Divisor2);

    Counter3 = Counter1 * Counter2;
    Divisor3 = Divisor1 * Divisor2;
};

static void AddFract
    ( int Counter1,  int Divisor1
    , int Counter2,  int Divisor2
    , int &Counter3, int &Divisor3
    )
{
    ReduceFract (Counter1, Divisor1);
    ReduceFract (Counter2, Divisor2);

    Counter3 = Counter1 * Divisor2 + Counter2 * Divisor1;
    Divisor3 = Divisor1 * Divisor2;
};

static void MultRate
    ( int Counter1,  int Divisor1
    , int Counter2,  int Divisor2
    , int &Counter3, int &Divisor3
    )
{
    int Ca, Da, Cb, Db;

    AddFract
        ( Counter1, Divisor1
        , Counter2, Divisor2
        , Ca,       Da
        );

    MultFract
        ( Counter1, Divisor1
        , Counter2, Divisor2
        , Cb,       Db
        );

    AddFract
        ( Ca,       Da
        , Cb,       Db
        , Counter3, Divisor3
        );
};

EzString NrSpeedBuf::GetSound (unsigned int MaxSize, long MsCurTime)
{
    EzString RetVal, HistSound;
    int      Counter, Divisor, TmpGetSpeed, CorrVal, DiffCount;
    int      tC, tD, tSC, tSD;
    int      ExpSize;

    GetCalc.AddData (MaxSize, MsCurTime);

    TmpGetSpeed = (GetCalc.GetAvgSpeed (8) * 1000);
    AdjustAvg (AvgGetSpeed, TmpGetSpeed, 6);

    CorrVal = GetBufUsed () - (GetBufSize () >> 1);

    Counter = AvgPutSpeed - AvgGetSpeed;
    Divisor = AvgPutSpeed;

    MultRate
        ( CorrVal,  1000000
        , Counter,  Divisor
        , tC,       tD
        );

    if (AvgDivisor == 0) return 0;

    tSC = tC;
    tSD = tD;

    if (tSC < -(tSD >> 2)) tSC = -(tSD >> 2);
    if (tSC >  (tSD >> 2)) tSC =  (tSD >> 2);

    tSC <<= 2;

    ReduceFract (tSC, tSD);

    AdjustAvg (AvgCounter, tSC, 2);
    AdjustAvg (AvgDivisor, tSD, 2);

    DiffCount = GetBufUsed ();

    if (State == On) {
        RetVal = NrFifo::GetSound
            ( MaxSize
            , tC, tD
            , SpeedCount
            );
    } else {
        RetVal = Rpad ("", MaxSize, 0);
    };

    DiffCount -= GetBufUsed ();
    ExpCount  += DiffCount;

    while (ExpCount > 0 && HistBuf.GetDataSize () > 0) {
        HistSound = HistBuf.GetData (ExpSize);
        HandleRecordData (HistSound, MsCurTime);
        ExpCount  -= ExpSize;
    };

    return RetVal;
};

int NrSpeedBuf::GetBufRate (void)
{
    return (100 * GetBufUsed ()) / (GetBufSize () >> 1);
};

void NrSpeedBuf::HandleRecordData (EzString RecData, long MsCurTime)
{
};

int NrSpeedBuf::GetAdjustRate (void)
{
    int tC, tD;

    if (AvgDivisor == 0) return 0;

    tC = AvgCounter << 8;
    tD = AvgDivisor;

    return tC / tD;
};

